<?php
/*
 * Class: WebRequest
 * ~~~~~~~~~~~~~~~~~
 * » Provides asynchronous and synchronous communication for HTTP GET-, POST- and HEAD-Requests.
 *
 * ----------------------------------------------------------------------------------
 *
 * LICENSE: This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ----------------------------------------------------------------------------------
 *
 */


/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

class WebRequest extends BaseClass {
	private $timeout;
	private $timeout_dns;
	private $timeout_connect;
	private $tries;
	private $max_redirect;

	private $spool;
	private $path;
	private $commands;


	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function __construct () {
		global $aseco;

		$this->setAuthor('undef.de');
		$this->setVersion('1.0.0');
		$this->setBuild('2017-04-23');
		$this->setCopyright('2016 - 2017 by undef.de');
		$this->setDescription('Provides asynchronous and synchronous communication for HTTP GET-, POST- and HEAD-Requests.');

		$this->timeout		= 20;
		$this->timeout_dns	= 5;
		$this->timeout_connect	= 5;
		$this->tries		= 10;
		$this->max_redirect	= 50;

		$this->spool		= array();
		$this->path		= 'cache/webrequest/';
		$this->commands		= array(
			'POSIX'		=> array(
				'prefix'	=> '',
				'postfix'	=> ' > /dev/null 2>&1 &',		// Start in background and ignore output
				'wget'		=> '/usr/bin/wget',
				'move'		=> 'mv --force',
			),
			'WINDOWS'		=> array(
				'prefix'	=> 'START /B ',				// Start in background
				'postfix'	=> ' >nul 2>&1',			// Ignore output
				'wget'		=> 'includes\wget.exe',
				'move'		=> 'move /Y',
			),
		);

		if (file_exists($this->commands[OPERATING_SYSTEM]['wget']) === false || is_file($this->commands[OPERATING_SYSTEM]['wget']) === false || is_executable($this->commands[OPERATING_SYSTEM]['wget']) === false) {
			$aseco->console('[WebRequest] This class requires a executable "wget"!');
			if (OPERATING_SYSTEM == 'WINDOWS') {
				$aseco->console('[WebRequest] Please download wget at "http://gnuwin32.sourceforge.net/packages/wget.htm" (Binaries as ZIP-File) and copy "bin\wget.exe" to "'. $this->commands[OPERATING_SYSTEM]['wget'] .'".');
			}
			else if (OPERATING_SYSTEM == 'POSIX') {
				$aseco->console('[WebRequest] Please install it with "sudo apt-get install wget" (or a similar command related to your distribution) and make sure it is accessible at "'. $this->commands[OPERATING_SYSTEM]['wget'] .'".');
			}
			die();
		}

		// Make sure the directory exists
		@mkdir($this->path, 0755, true);
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function HEAD ($params) {

		// Setup method
		$params['method'] = 'HEAD';

		if (!isset($params['url']) || empty($params['url'])) {
			throw new WebRequestException('No required URL given for a request!', WebRequestException::NO_URL_GIVEN_FOR_REQUEST);
		}
		if (!isset($params['user_agent']) || empty($params['user_agent'])) {
			$params['user_agent'] = get_class($this) .'/'. $this->getVersion() .' '. USER_AGENT;
		}
		if (!isset($params['timeout']) || empty($params['timeout'])) {
			$params['timeout'] = $this->timeout;
		}
		if (!isset($params['timeout_dns']) || empty($params['timeout_dns'])) {
			$params['timeout_dns'] = $this->timeout_dns;
		}
		if (!isset($params['timeout_connect']) || empty($params['timeout_connect'])) {
			$params['timeout_connect'] = $this->timeout_connect;
		}
		if (!isset($params['tries']) || empty($params['tries'])) {
			$params['tries'] = $this->tries;
		}
		if (!isset($params['max_redirect']) || empty($params['max_redirect'])) {
			$params['max_redirect'] = $this->max_redirect;
		}
		if (!isset($params['extra_headers']) || empty($params['extra_headers'])) {
			$params['extra_headers'] = array();
		}

		$stream_context = stream_context_create(
			array(
				'ssl'		=> array(
					'verify_peer'		=> false,
					'verify_peer_name'	=> false,
					'allow_self_signed'	=> true,
					'SNI_enabled'		=> true,
				),
				'http'		=> array(
					'ignore_errors'		=> false,
					'timeout'		=> $params['timeout'],
					'follow_location'	=> true,
					'max_redirects'		=> $params['max_redirect'],
					'protocol_version'	=> 1.1,
					'user_agent'		=> $params['user_agent'],
					'method'		=> $params['method'],
					'header'		=> $params['extra_headers'],
				),
			)
		);

		// Send sync request
		$response = file_get_contents($params['url'], false, $stream_context);

		$headers = '';
		foreach ($http_response_header as $line) {
			$headers .= $line ."\r\n";
		}

		// Create new construct
		$request = new WebRequestConstruct($params);

		// Parse the response
		return $this->_parse($request, $headers, $response);
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function GET ($params) {

		// Setup method
		$params['method'] = 'GET';

		if (!isset($params['url']) || empty($params['url'])) {
			throw new WebRequestException('No required URL given for a request!', WebRequestException::NO_URL_GIVEN_FOR_REQUEST);
		}
		if (!isset($params['user_agent']) || empty($params['user_agent'])) {
			$params['user_agent'] = get_class($this) .'/'. $this->getVersion() .' '. USER_AGENT;
		}
		if (!isset($params['timeout']) || empty($params['timeout'])) {
			$params['timeout'] = $this->timeout;
		}
		if (!isset($params['timeout_dns']) || empty($params['timeout_dns'])) {
			$params['timeout_dns'] = $this->timeout_dns;
		}
		if (!isset($params['timeout_connect']) || empty($params['timeout_connect'])) {
			$params['timeout_connect'] = $this->timeout_connect;
		}
		if (!isset($params['tries']) || empty($params['tries'])) {
			$params['tries'] = $this->tries;
		}
		if (!isset($params['max_redirect']) || empty($params['max_redirect'])) {
			$params['max_redirect'] = $this->max_redirect;
		}
		if (!isset($params['extra_headers']) || empty($params['extra_headers'])) {
			$params['extra_headers'] = array();
		}

		if (isset($params['sync']) && $params['sync'] === true) {
			$stream_context = stream_context_create(
				array(
					'ssl'		=> array(
						'verify_peer'		=> false,
						'verify_peer_name'	=> false,
						'allow_self_signed'	=> true,
						'SNI_enabled'		=> true,
					),
					'http'		=> array(
						'ignore_errors'		=> false,
						'timeout'		=> $params['timeout'],
						'follow_location'	=> true,
						'max_redirects'		=> $params['max_redirect'],
						'protocol_version'	=> 1.1,
						'user_agent'		=> $params['user_agent'],
						'method'		=> $params['method'],
						'header'		=> $params['extra_headers'],
					),
				)
			);

			// Send sync request
			$response = file_get_contents($params['url'], false, $stream_context);

			$headers = '';
			foreach ($http_response_header as $line) {
				$headers .= $line ."\r\n";
			}

			// Create new construct
			$request = new WebRequestConstruct($params);

			// Parse the response
			return $this->_parse($request, $headers, $response);
		}
		else {
			if (!isset($params['callback']) && $params['callback'] !== null) {
				throw new WebRequestException('No required callback given for a asynchronos request!', WebRequestException::NO_CALLBACK_GIVEN_ON_ASYNC_REQUEST);
			}
			else if ($params['callback'] !== null && (!is_callable($params['callback']) && !is_callable($params['callback'][0]))) {
				throw new WebRequestException('Given callback is not callable!', WebRequestException::GIVEN_CALLBACK_IS_NOT_CALLABLE);
			}

			// Create new construct
			$request = new WebRequestConstruct($params);

			// Add to the async pool
			$this->spool[$request->id] = $request;
			return $request->id;
		}
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function POST ($params) {

		// Setup method
		$params['method'] = 'POST';

		if (!isset($params['url']) || empty($params['url'])) {
			throw new WebRequestException('No required URL given for a request!', WebRequestException::NO_URL_GIVEN_FOR_REQUEST);
		}
		if (!isset($params['data']) || empty($params['data'])) {
			throw new WebRequestException('No required POST data given!', WebRequestException::NO_POST_DATA_GIVEN);
		}
		if (!isset($params['user_agent']) || empty($params['user_agent'])) {
			$params['user_agent'] = get_class($this) .'/'. $this->getVersion() .' '. USER_AGENT;
		}
		if (!isset($params['timeout']) || empty($params['timeout'])) {
			$params['timeout'] = $this->timeout;
		}
		if (!isset($params['timeout_dns']) || empty($params['timeout_dns'])) {
			$params['timeout_dns'] = $this->timeout_dns;
		}
		if (!isset($params['timeout_connect']) || empty($params['timeout_connect'])) {
			$params['timeout_connect'] = $this->timeout_connect;
		}
		if (!isset($params['tries']) || empty($params['tries'])) {
			$params['tries'] = $this->tries;
		}
		if (!isset($params['max_redirect']) || empty($params['max_redirect'])) {
			$params['max_redirect'] = $this->max_redirect;
		}
		if (!isset($params['extra_headers']) || empty($params['extra_headers'])) {
			$params['extra_headers'] = array();
		}

		if (isset($params['sync']) && $params['sync'] === true) {
			$params['data'] = http_build_query($params['data']);
			$stream_context = stream_context_create(
				array(
					'ssl'		=> array(
						'verify_peer'		=> false,
						'verify_peer_name'	=> false,
						'allow_self_signed'	=> true,
						'SNI_enabled'		=> true,
					),
					'http'		=> array(
						'ignore_errors'		=> false,
						'timeout'		=> $params['timeout'],
						'follow_location'	=> true,
						'max_redirects'		=> $params['max_redirect'],
						'protocol_version'	=> 1.1,
						'user_agent'		=> $params['user_agent'],
						'method'		=> $params['method'],
						'header'		=> array_merge(
							array(
								'Content-Type: application/x-www-form-urlencoded',
								'Connection: close',
								'Content-Length: '. strlen($params['data']),
							),
							$params['extra_headers']
						),
						'content'		=> $params['data'],
					),
				)
			);

			// Send sync request
			$response = file_get_contents($params['url'], false, $stream_context);

			$headers = '';
			foreach ($http_response_header as $line) {
				$headers .= $line ."\r\n";
			}

			// Create new construct
			$request = new WebRequestConstruct($params);

			// Parse the response
			return $this->_parse($request, $headers, $response);
		}
		else {
			if (!isset($params['callback']) && $params['callback'] !== null) {
				throw new WebRequestException('No required callback given for a asynchronos request!', WebRequestException::NO_CALLBACK_GIVEN_ON_ASYNC_REQUEST);
			}
			else if ($params['callback'] !== null && (!is_callable($params['callback']) && !is_callable($params['callback'][0]))) {
				throw new WebRequestException('Given callback is not callable!', WebRequestException::GIVEN_CALLBACK_IS_NOT_CALLABLE);
			}

			// Create new construct
			$request = new WebRequestConstruct($params);

			// Add to the async pool
			$this->spool[$request->id] = $request;
			return $request->id;
		}
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	private function _async_get ($request) {

		$extra_headers[] = '--header="X-WebRequest-ID: '. $request->id .'" \\';
		if (count($request->extra_headers) > 0) {
			foreach ($request->extra_headers as $header) {
				$extra_headers[] = '--header="'. $header .'" \\';
			}
		}

		$command = '
		(
			'. $this->commands[OPERATING_SYSTEM]['prefix'] .' '. $this->commands[OPERATING_SYSTEM]['wget'] .' \
				--quiet \
				--no-verbose \
				--continue \
				--retry-connrefused \
				--timeout='. $request->timeout .' \
				--dns-timeout='. $request->timeout_dns .' \
				--connect-timeout='. $request->timeout_connect .' \
				--tries='. $request->tries .' \
				--max-redirect='. $request->max_redirect .' \
				--save-headers \
				--no-cache \
				--no-http-keep-alive \
				--no-directories \
				--no-cookies \
				--no-check-certificate \
				--method=GET \
				'. implode("\n", $extra_headers) .'
				--user-agent="'. $request->user_agent .'" \
				--output-document="'. $this->path . $request->id .'.part" \
				"'. $request->url .'" &&
			'. $this->commands[OPERATING_SYSTEM]['move'] .' \
				"'. $this->path . $request->id .'.part" \
				"'. $this->path . $request->id .'.done"
		) '. $this->commands[OPERATING_SYSTEM]['postfix'] .'
		';

		// Clean up the $command
		$command = preg_replace('/[ \t]{2,}/', ' ', $command);
		$command = preg_replace('/\n{2,}/', "\n", $command);

		// We do not care about the output/results here, just execute into background
		exec($command);
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	private function _async_post ($request) {

		$extra_headers[] = '--header="X-WebRequest-ID: '. $request->id .'" \\';
		if (count($request->extra_headers) > 0) {
			foreach ($request->extra_headers as $header) {
				$extra_headers[] = '--header="'. $header .'" \\';
			}
		}

		$command = '
		(
			'. $this->commands[OPERATING_SYSTEM]['prefix'] .' '. $this->commands[OPERATING_SYSTEM]['wget'] .' \
				--quiet \
				--no-verbose \
				--continue \
				--retry-connrefused \
				--timeout='. $request->timeout .' \
				--dns-timeout='. $request->timeout_dns .' \
				--connect-timeout='. $request->timeout_connect .' \
				--tries='. $request->tries .' \
				--max-redirect='. $request->max_redirect .' \
				--save-headers \
				--no-cache \
				--no-http-keep-alive \
				--no-directories \
				--no-cookies \
				--no-check-certificate \
				--method=POST \
				'. implode("\n", $extra_headers) .'
				--user-agent="'. $request->user_agent .'" \
				--post-data="'. $request->data .'" \
				--output-document="'. $this->path . $request->id .'.part" \
				"'. $request->url .'" &&
			'. $this->commands[OPERATING_SYSTEM]['move'] .' \
				"'. $this->path . $request->id .'.part" \
				"'. $this->path . $request->id .'.done"
		) '. $this->commands[OPERATING_SYSTEM]['postfix'] .'
		';

		// Clean up the $command
		$command = preg_replace('/[ \t]{2,}/', ' ', $command);
		$command = preg_replace('/\n{2,}/', "\n", $command);

		// We do not care about the output/results here, just execute into background
		exec($command);
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function update () {

		if (count($this->spool) > 0) {
			foreach ($this->spool as $id => $request) {
				if ($request->sent === false) {
					if ($request->method == 'GET') {
						$this->_async_get($request);
					}
					else if ($request->method == 'POST') {
						$this->_async_post($request);
					}
					$request->sent = true;
				}
				else if ($request->sent === true) {
					if (file_exists($this->path . $request->id .'.done')) {
						$content = file_get_contents($this->path . $request->id .'.done');
						unlink($this->path . $request->id .'.done');

						if ($request->callback !== null) {
							// Parse the response
							$request = $this->_parse($request, $content, false);

							// Call the callback function
							if (is_callable($request->callback)) {
								call_user_func($request->callback, $request);
							}
							else if (is_callable($request->callback[0])) {
								call_user_func($request->callback[0], $request, $request->callback[1]);
							}
						}

						// Destroy the request
						$request = null;
						unset($this->spool[$id]);
					}
				}
			}
		}

		// Cleanup
		if ($dir = opendir($this->path)) {
			while ($entry = readdir($dir)) {
				if (@is_file($this->path.DIRECTORY_SEPARATOR.$entry) === true && strpos('.done', $entry) !== false) {
					if (@filemtime($this->path.DIRECTORY_SEPARATOR.$entry) + 86400 < time()) {
						@unlink($this->path.DIRECTORY_SEPARATOR.$entry);
					}
				}
			}
		}
	}

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	private function _parse ($request, $header, $content = false) {

		if ($content === false) {
			// wget
			$content = str_replace("\r", '', $header);	// DOS to UNIX
			$data = explode("\n\n", $content);		// Split header and content
			$request->response['content'] = $data[1];
			$data[1] = null;				// Purge memory
		}
		else {
			// file_get_contents()
			$data[0] = $header;
			$request->response['content'] = $content;
			$content = null;				// Purge memory
		}

		// Parse headers
		foreach (explode("\n", $data[0]) as $line) {
			if (substr($line, 0, 5) == 'HTTP/' || substr($line, 0, 6) == 'HTTPS/') {
				$parts = explode(' ', $line);
				$request->response['header']['protocol']	= $parts[0];
				$request->response['header']['code']		= (int)$parts[1];
				$request->response['header']['code_text']	= $parts[2];
			}
			else {
				$item = explode(': ', $line);
				if (strtolower($item[0]) == 'last-modified') {
					$request->response['header']['last_modified'] = $item[1];
				}
				else if (strtolower($item[0]) == 'content-type') {
					$request->response['header']['content_type'] = $item[1];
				}
			}
		}
		return $request;
	}
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

class WebRequestConstruct {
	public $timeout			= 20;
	public $timeout_dns		= 5;
	public $timeout_connect		= 5;
	public $tries			= 10;
	public $max_redirect		= 50;

	public $id			= false;
	public $timestamp		= 0;
	public $method			= 'GET';
	public $url			= false;
	public $extra_headers		= array();
	public $data			= false;
	public $user_agent		= false;
	public $sent			= false;
	public $callback		= false;

	public $response		= array(
		'header'	 => array(
			'protocol'		=> '',
			'code'			=> 0,
			'code_text'		=> '',
			'last_modified'		=> 0,
			'content_type'		=> false,
		),
		'content'	=> false,
	);

	/*
	#///////////////////////////////////////////////////////////////////////#
	#									#
	#///////////////////////////////////////////////////////////////////////#
	*/

	public function __construct ($params) {
		$this->timeout		= $params['timeout'];
		$this->timeout_dns	= $params['timeout_dns'];
		$this->timeout_connect	= $params['timeout_connect'];
		$this->tries		= $params['tries'];
		$this->max_redirect	= $params['max_redirect'];

		$this->id		= md5($params['method'] . $params['url'] . microtime(false));
		$this->timestamp	= microtime(false);
		$this->method		= strtoupper($params['method']);
		$this->url		= $params['url'];
		$this->extra_headers	= (isset($params['extra_headers']) ? $params['extra_headers'] : array());
		$this->data		= (isset($params['data']) ? $params['data'] : false);
		$this->callback		= (isset($params['callback']) ? $params['callback'] : false);
		$this->user_agent	= $params['user_agent'];
	}
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

class WebRequestException extends Exception {
	const NO_URL_GIVEN_FOR_REQUEST			= 501;
	const NO_POST_DATA_GIVEN			= 502;
	const NO_CALLBACK_GIVEN_ON_ASYNC_REQUEST	= 503;
	const GIVEN_CALLBACK_IS_NOT_CALLABLE		= 504;
}

?>
